
#define OFFSET 80

#include "game.h"
#include <stdio.h>

typedef enum { s_init,
    s_create_new, s_verify_creation, s_play, s_drop,
    s_add_block, s_remove_lines, s_remove_1, s_remove_2,
    s_game_over } Engine_State;


static Engine_State run_state = s_init;
static Tet_State old_tet_state;
static BYTE player_move_counter;
static BYTE remove_line;
static BYTE tet_bag;

#define FULL_BAG ((1 << FIG_NUM) - 1)

DWORD packed_well[NUM_ROWS];
Tet_State tet_state;
Tet_State incoming[NUM_INCOMING];
BYTE player_moves_per_down_move;
WORD score, level;
DWORD rng_value;
extern DWORD arng_value;

static BYTE rng(void)
{
    rng_value = ((((DWORD) rng_value & (DWORD) 0xffffL) * (DWORD) 36969L) 
            + ((DWORD) rng_value >> (DWORD) 16L));
    return (BYTE) rng_value;
}

extern void halt_and_catch_fire(BYTE info);

#define init_state(d)       \
    d.x = (NUM_COLUMNS / 2) - 1;\
    d.y = -NUM_ROWS;        \
    d.r = 0;                \
    d.i = 0;

#define copy_state(d, s)    \
    d.x = s.x;              \
    d.y = s.y;              \
    d.r = s.r;              \
    d.i = s.i;

static BYTE collision_detect(void)
{
    BYTE i;
    for (i = 0; i < SQUARE_NUM; i++) {
        char x = TETRITABLE[tet_state.i][tet_state.r][i][0] + tet_state.x;
        char y = TETRITABLE[tet_state.i][tet_state.r][i][1] + tet_state.y;

        if ((y >= NUM_ROWS) || (x < 0) || (x >= NUM_COLUMNS)) {
            return 1;
        }
        if (y < 0) {
            continue;
        }
        if (get_well(y, x) == 1) {
            return 1;
        }
    }
    return 0;
}

static void no_collisions(BYTE info)
{
    if (collision_detect()) {
        halt_and_catch_fire(info);
    }
}

static BYTE add_remove(BYTE add)
{
    BYTE i, off_the_top = 0;
    no_collisions(4);
    for (i = 0; i < SQUARE_NUM; i++) {
        char x = TETRITABLE[tet_state.i][tet_state.r][i][0] + tet_state.x;
        char y = TETRITABLE[tet_state.i][tet_state.r][i][1] + tet_state.y;
        DWORD shift;

        if (y < 0) {
            off_the_top = 1;
            continue;
        } 
        if ((y >= NUM_ROWS) || (x < 0) || (x >= NUM_COLUMNS)) {
            continue;
        }
        shift = 2 * x;
        packed_well[(BYTE) (y)] &= ~((DWORD) 3 << shift);
        packed_well[(BYTE) (y)] |= ((DWORD) add << shift);
    }
    return off_the_top;
}

static BYTE full_line_detect(void)
{
    BYTE x, y;
    y = NUM_ROWS;
    do {
        BYTE line_full = 1;
        y --;
        for (x = 0; x < NUM_COLUMNS; x++) {
            char v = get_well(y, x);
            if (v == 0) { 
                line_full = 0;
            } else if (v != 1) {
                halt_and_catch_fire(11);
            }
        }
        if (line_full) {
            return y;
        }
    } while (y != 0);

    return NUM_ROWS;
}

static void generate_tet(Tet_State * s)
{
    BYTE mask = 0;

    init_state((*s));

    // Bag empty?
    if (tet_bag == 0) {
        tet_bag = FULL_BAG;
    }

    // Remove one tetromino from the bag
    do {
        s->i = rng() & 7;
        mask = 1 << s->i;
    } while ((mask & tet_bag) == 0);

    //tet_bag &= ~mask;

    // Pick rotation
    s->r = rng() & ROTATIONS[s->i];
}

static void set_level_for_score(void)
{
    WORD l = score / 10;

    if (l >= level) level = l;
    if (level > 99) level = 99;
    if (level < 1) level = 1;

    l = level / 5;
    if (l < 5) {
        player_moves_per_down_move = 5 - l;
    } else {
        player_moves_per_down_move = 0;
    }
}


void game_reset(void)
{
    BYTE y, i;
    for (y = 0; y < NUM_ROWS; y++) {
        packed_well[y] = 0;
    }
    run_state = s_init;
    remove_line = 0;
    init_state(old_tet_state);
    init_state(tet_state);
    tet_bag = FULL_BAG;
    level = score = 0;
    set_level_for_score();

    for (i = 0; i < NUM_INCOMING; i++) {
        generate_tet(&incoming[i]);
    }
}

void game_init(WORD number)
{
    rng_value = (DWORD) number;
    rng_value += OFFSET;
    arng_value = rng_value;
}

BYTE game_delete(void)
{
    set_level_for_score();

    if ((run_state == s_play) 
    || (run_state == s_add_block) 
    || (run_state == s_drop)) {
        add_remove(0); // remove the tetrimino
        run_state = s_create_new;   // and create another
        return 1;
    }
    return 0;
}

Game_State game_step(void)
{
    BYTE i;

    add_remove(0); /* deleted */ 
    no_collisions(7);
    switch (run_state) {
    case s_init:
        game_reset();
        run_state = s_create_new;
        break;
    case s_create_new:
        init_state(old_tet_state);
        copy_state(tet_state, incoming[0]);
        tet_state.y = -2;
        tet_state.x = (NUM_COLUMNS / 2) - 1;

        for (i = 1; i < NUM_INCOMING; i++) {
            copy_state(incoming[i-1], incoming[i]);
        }
        generate_tet(&incoming[NUM_INCOMING - 1]);

        if (collision_detect()) {
            run_state = s_game_over;
        } else {
            run_state = s_play;
        }
        player_move_counter = player_moves_per_down_move;
        break;
    case s_play:
        no_collisions(2);

        if (player_move_counter == 0) {
            copy_state(old_tet_state, tet_state);
            player_move_counter = player_moves_per_down_move;
            tet_state.y ++;
            if (collision_detect()) {
                // Tetrimino collides due to falling; frozen in place
                copy_state(tet_state, old_tet_state);
                no_collisions(3);
                run_state = s_add_block;
            }
        } else {
            player_move_counter --;
        }
        if (run_state == s_play) {
            Player_Move pmove = ai_move();

            copy_state(old_tet_state, tet_state);

            switch(pmove) {
            case MOVE_LEFT:     tet_state.x --; break;
            case MOVE_RIGHT:    tet_state.x ++; break;
            case MOVE_ROTATE:   tet_state.r ++; 
                                tet_state.r &= ROTATIONS[tet_state.i]; break;
            case MOVE_DROP:     tet_state.y ++; 
                                run_state = s_drop; break;
            default:            break;
            }
            if (collision_detect()) {
                // invalid move, undo 
                if (pmove == MOVE_ROTATE) {
                    // try "wall kick" left
                    tet_state.x --;
                    if (collision_detect()) {
                        // didn't work, try "wall kick" right
                        tet_state.x += 2;
                        if (collision_detect()) {
                            // Still no good, undo
                            copy_state(tet_state, old_tet_state);
                        }
                    }
                } else {
                    // Always undo
                    copy_state(tet_state, old_tet_state);
                }
            }
        }
        no_collisions(1);
        break;
    case s_drop:
        if (level > (NUM_ROWS / 3)) {
            i = NUM_ROWS / 3;
        } else {
            i = level + 1;
        }
        while (i) {
            i--;
            copy_state(old_tet_state, tet_state);
            tet_state.y ++;
            if (collision_detect()) {
                // undo drop and finish up
                copy_state(tet_state, old_tet_state);
                no_collisions(92);
                run_state = s_add_block;
                break;
            }
        }
        break;
    case s_add_block:
        run_state = s_remove_lines;
        no_collisions(6);
        if (add_remove(1)) {
            run_state = s_game_over;
        }
        init_state(tet_state);
        break;
    case s_remove_lines:
        // Remove a full line
        i = full_line_detect();
        if (i < NUM_ROWS) {
            run_state = s_remove_1;
            remove_line = i;
            // score & level updated
            score ++;
            set_level_for_score();
        } else {
            run_state = s_create_new; // nothing found
        }
        break;
    case s_remove_1:
        run_state = s_remove_2;
        i = remove_line;
        if (i < NUM_ROWS) {
            packed_well[i] = 0;
        }
        break;
    case s_remove_2:
        run_state = s_remove_lines;
        i = remove_line;
        if (i < NUM_ROWS) {
            BYTE y;
            for (y = i; y >= 1; y--) {
                packed_well[y] = packed_well[y-1];
            }
            packed_well[0] = 0;
        }
        break;
    default: // (s_game_over)
        // Await reset
        run_state = s_game_over;
        break;
    }
    no_collisions(5);
    add_remove(2); /* added (temporary) */ 

    switch(run_state) {
    case s_game_over:   return GAME_OVER;
    case s_remove_2:    return INC_SPEED;
    default:            return GAME_CONTINUE;
    }
}
    

